/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <json-c/json.h>

#include "common.h"
#include "faultd-config.h"
#include "log.h"
#include "module.h"

static int parse_disable_modules(struct json_object *root)
{
	struct json_object *node;
	const char *name;
	struct faultd_module *module;
	int i;
	int len;

	len = json_object_array_length(root);

	for (i = 0; i < len; ++i) {
		node = json_object_array_get_idx(root, i);
		if (!node)
			return -EINVAL;

		name = json_object_get_string(node);
		if (!name)
			return -EINVAL;

		module = faultd_get_module_by_name(name);
		if (module) {
			module->disabled = 1;
			log_debug("Disabled %s module", name);
		} else {
			log_warning("Could not find %s module", name);
		}
	}

	return 0;
}

int faultd_parse_config(const char *config_path, struct faultd_config *c)
{
	struct json_object *root;
	struct json_object *node;
	int ret = 0;
	json_bool json_ret;

	root = json_object_from_file(config_path);
	if (is_error(root)) {
		log_error("Could not create object from json file %s", config_path);
		return -EINVAL;
	}

	json_ret = json_object_object_get_ex(root, "disable_modules", &node);
	if (json_ret) {
		ret = parse_disable_modules(node);
		if (ret < 0) {
			log_error("Could not parse configuration file %s", config_path);
			goto out;
		}
	}

	c->root = root;
	return 0;

out:
	json_object_put(root);
	return ret;
}

void faultd_config_cleanup(struct faultd_config *c)
{
	if (c)
		json_object_put(c->root);
}

int faultd_config_get_subconf(struct faultd_config *conf, const char *key, struct faultd_config **sub)
{
	struct faultd_config *config_node;
	struct json_object *node;
	json_bool ret;

	ret = json_object_object_get_ex(conf->root, key, &node);
	if (!ret)
		return -ENOENT;

	config_node = calloc(1, sizeof(*config_node));
	if (!config_node)
		return -ENOMEM;

	if (!json_object_is_type(node, json_type_object)) {
		log_error("%s has to be an object", key);
		goto out;
	}

	config_node->root = json_object_get(node);

	*sub = config_node;
	return 0;

out:
	free(config_node);
	return -EINVAL;
}
