/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <malloc.h>

#include "service_failed_event.h"

static int allocate_sf_event(struct faultd_event_type *type,
                             void *data, struct faultd_event **ev)
{
	struct service_failed_event *sf_ev;
	struct sf_event_data *sf_ev_data = data;
	int ret;

	sf_ev = calloc(1, sizeof(*sf_ev));
	if (!sf_ev)
		return -ENOMEM;

	ret = faultd_event_init_internal(type, &sf_ev->event);
	if (ret)
		goto free_sf_ev;

	/* TODO */
	sf_ev->service = sf_ev_data->service;
	sf_ev->detection_time = sf_ev_data->detection_time;
	/* /TODO */

	*ev = &sf_ev->event;
	return 0;
free_sf_ev:
	free(sf_ev);

	return ret;
}

static int deserialize_sf_event(struct faultd_event_type *type,
								struct faultd_object *data, struct faultd_event **ev)
{
	int ret = -EINVAL;
	struct sf_event_data sf_ev_data;
	struct faultd_object *obj;

	memset(&sf_ev_data, 0, sizeof(sf_ev_data));

	list_for_each_entry(obj, &data->val.children, node) {
		if ((obj->type == TYPE_OBJECT) &&
			(strcmp(SYSD_SERVICE, obj->key) == 0)) {

			ret = systemd_service_deserialize(obj, &sf_ev_data.service);
			if (ret < 0)
				goto finish;
		} else if ((obj->type == TYPE_TIME_T) &&
				   (strcmp(SF_EV_DTIME, obj->key) == 0)) {
			sf_ev_data.detection_time = obj->val.time;
		}
	}

	ret = allocate_sf_event(type, &sf_ev_data, ev);
	if (ret < 0) {
		systemd_service_cleanup(&sf_ev_data.service);
		goto finish;
	}

	ret = faultd_event_deserialize_internal(data, type, *ev);
	if (ret < 0) {
		struct service_failed_event *sf_ev =
			to_service_failed_event(*ev);
		free(sf_ev_data.service.dbus_path);
		sf_ev_data.service.dbus_path = NULL;
		free(sf_ev);
		goto finish;
	}

	ret = 0;
finish:
	return ret;
}

static void sf_event_release(struct faultd_event *ev)
{
	struct service_failed_event *sf_ev =
		to_service_failed_event(ev);

	systemd_service_cleanup(&sf_ev->service);
	faultd_event_cleanup_internal(&sf_ev->event);
	free(sf_ev);
}

static char *sf_event_to_string(struct faultd_event *ev)
{
	struct service_failed_event *sf_ev =
		to_service_failed_event(ev);
	char *str;
	int ret;

	/* TODO print other fields */
	ret = asprintf(&str, "Service Failed Event:"
				   " Time: %lld.%.9ld"
				   " Service: path %s",
				   (long long)sf_ev->event.timestamp.tv_sec,
				   sf_ev->event.timestamp.tv_nsec,
				   sf_ev->service.dbus_path);

	return ret > 0 ? str : NULL;
}

static void sf_event_serialize(struct faultd_event *ev, struct faultd_object *out)
{
	struct service_failed_event *sf_ev =
		to_service_failed_event(ev);
	faultd_event_serialize_internal(ev, out);
	systemd_service_serialize(&sf_ev->service, out);
	faultd_object_append_time_t(out, SF_EV_DTIME, sf_ev->detection_time);
}
TEST_ENTRY_POINT(sf_event_serialize);

static struct faultd_event_type service_failed_event_type = {
	.name = SERVICE_FAILED_EVENT_ID,
	.default_ops = {
		.release = sf_event_release,
		.serialize = sf_event_serialize,
		.to_string = sf_event_to_string,
	},
	.allocate_event = allocate_sf_event,
	.deserialize_event = deserialize_sf_event,
	.node = LIST_HEAD_INIT(service_failed_event_type.node),
};

FAULTD_EVENT_TYPE_REGISTER(service_failed_event_type, service_failed_et)
